unit uQCScraper;

interface

uses
  System.Classes, ICDSReportPublicInterface1;

type
  TProgressEvent = reference to procedure(Value: Integer);
  TDataTypeEvent = reference to procedure(const Value: string);

  TQCScraper = class
  private
    FDataTypes: TArray<string>;
    FFileName: string;
    FIntf: ICDSReportPublicInterface;
    FOnDataType: TDataTypeEvent;
    FOnProgress: TProgressEvent;
    FPassword: string;
    FSessionID: string;
    FUsername: string;
    function GetDataTypes: TArray<string>;
    function GetIntf: ICDSReportPublicInterface;
    function GetSessionID: string;
  protected
    function CountData(const DataType: string): Integer;
    procedure DoDataType(const Value: string);
    procedure DoProgress(Value: Integer);
    function ExtractFirstRows(const S: string): string;
    function ExtractRowData(const S: string): string;
    function FetchData(const DataType: string; MaxCount, StartRow: Integer): string;
    procedure LoadAllData;
    procedure LoadData(Writer: TStreamWriter; const DataType: string);
    property DataTypes: TArray<string> read GetDataTypes;
    property Intf: ICDSReportPublicInterface read GetIntf;
    property SessionID: string read GetSessionID;
  public
    constructor Create(const AUsername, APassword, AFileName: string);
    procedure Execute;
    property FileName: string read FFileName;
    property Password: string read FPassword;
    property Username: string read FUsername;
    property OnDataType: TDataTypeEvent read FOnDataType write FOnDataType;
    property OnProgress: TProgressEvent read FOnProgress write FOnProgress;
  end;

implementation

uses
  System.SysUtils, System.Types, System.Generics.Collections, System.Zlib, System.NetEncoding,
  Datasnap.DBClient, MidasLib;

type
  TByteDynArrayHelper = record helper for TByteDynArray
  public
    function DecodeAndUnCompress: string;
  end;

function TByteDynArrayHelper.DecodeAndUnCompress: string;
var
  bytes: TBytes;
  fWorkStream: TMemoryStream;
  fEncodedStream: TMemoryStream;
  fStreamDeCompressor: TDeCompressionStream;
begin
  fEncodedStream := TMemoryStream.create;
  try
    fWorkStream := TMemoryStream.Create;
    try
      fWorkStream.Write(Self[0], Length(Self));
      fWorkStream.Position := 0;
      TNetEncoding.Base64.Decode(fWorkStream, fEncodedStream);
    finally
      fWorkStream.Free;
    end;
    fEncodedStream.position := 0;
    fWorkStream := TMemoryStream.create;
    try
      fStreamDeCompressor := TDeCompressionStream.create(fEncodedStream);
      try
        fWorkStream.CopyFrom(fStreamDeCompressor, 0);
      finally
        FreeAndNil(fStreamDeCompressor);
      end;
      fWorkStream.position := 0;
      SetLength(bytes, fWorkStream.size);
      fWorkStream.Read(bytes, fWorkStream.size);
    finally
      fWorkStream.Free;
    end;
  finally
    FreeAndnil(fEncodedStream);
  end;
  Result := TEncoding.ANSI.GetString(bytes);
end;

constructor TQCScraper.Create(const AUsername, APassword, AFileName: string);
begin
  inherited Create;
  FUsername := AUsername;
  FPassword := APassword;
  FFileName := AFileName;
end;

function TQCScraper.CountData(const DataType: string): Integer;
begin
  Result := intf.SearchCount(
      SessionID,
      '', //EditFirst.Text, // First_Name
      '', //EditLast.Text, // Last_Name
      '', // Defect_No
      '', // RAID_No
      '', // Status
      DataType, // Data_Type
      '', // Severity
      '', // Resolution
      '', // Project
      '', // Area
      '', // Version
      '', // keywords
      False, // workarounds
      '', // Build_no
      '', // Resolved_in_Build
      nil, // Since
      '', // ThePlatform
      '', // IgnoreAutoRep
      nil, // S_Date_Reported
      nil, // E_Date_Reported
      nil, // S_Modified_Date
      nil // E_Modified_Date
      );
end;

procedure TQCScraper.DoDataType(const Value: string);
begin
  if Assigned(FOnDataType) then FOnDataType(Value);
end;

procedure TQCScraper.DoProgress(Value: Integer);
begin
  if Assigned(FOnProgress) then FOnProgress(Value);
end;

function TQCScraper.ExtractFirstRows(const S: string): string;
var
  I: Integer;
begin
  I := S.IndexOf('</ROWDATA>');
  Result := S.Substring(0, I);
end;

function TQCScraper.ExtractRowData(const S: string): string;
var
  I: Integer;
begin
  I := S.IndexOf('<ROWDATA>');
  Result := ExtractFirstRows(S.Substring(I + 9));
end;

function TQCScraper.FetchData(const DataType: string; MaxCount, StartRow: Integer): string;
begin
  Result := intf.Search(
      SessionID,
      '', //EditFirst.Text, // First_Name
      '', //EditLast.Text, // Last_Name
      '', // Defect_No
      '', // RAID_No
      '', // Status
      DataType, // Data_Type
      '', // Severity
      '', // Resolution
      '', // Project
      '', // Area
      '', // Version
      '', // keywords
      False, // workarounds
      '', // Build_no
      '', // Resolved_in_Build
      nil, // Since
      MaxCount, // max_return
      '', // ThePlatform
      '', // IgnoreAutoRep
      StartRow, //Start_Row
      nil, // S_Date_Reported
      nil, // E_Date_Reported
      nil, // S_Modified_Date
      nil // E_Modified_Date
      ).DecodeAndUncompress;
end;

function TQCScraper.GetDataTypes: TArray<string>;
var
  cds: TClientDataSet;
  lst: TList<string>;
begin
  if FDataTypes = nil then begin
    cds := TClientDataSet.Create(nil);
    try
      cds.XMLData := intf.GetLookup(SessionID, 'Type').DecodeAndUncompress;
      lst := TList<string>.Create;
      try
        cds.First;
        while not cds.Eof do begin
          lst.Add(cds.FieldByName('LOOKUP_VALUE').AsString);
          cds.Next;
        end;
        FDataTypes := lst.ToArray;
      finally
        lst.Free;
      end;
    finally
      cds.Free;
    end;
  end;
  Result := FDataTypes;
end;

function TQCScraper.GetIntf: ICDSReportPublicInterface;
begin
  if FIntf = nil then begin
    FIntf := GetICDSReportPublicInterface;
  end;
  result := FIntf;
end;

function TQCScraper.GetSessionID: string;
var
  cds: TClientDataSet;
begin
  if FSessionID = '' then
  begin
    cds := TClientDataSet.Create(nil);
    try
      cds.XMLData := intf.Login(Username, Password, '', '').DecodeAndUncompress;
      FSessionID := cds.FieldByName('SESSIONID').AsString;
    finally
      cds.Free;
    end;
  end;
  Result := FSessionID;
end;

procedure TQCScraper.LoadAllData;
var
  S: string;
  writer: TStreamWriter;
begin
  writer := TStreamWriter.Create(FileName);
  try
    for S in DataTypes do begin
      DoDataType(S);
      LoadData(writer, S);
    end;
    DoDataType('');
    writer.WriteLine('</ROWDATA></DATAPACKET>');
  finally
    writer.Free;
  end;
end;

procedure TQCScraper.LoadData(Writer: TStreamWriter; const DataType: string);
const
  cMaxCount = 100;
var
  cnt: Integer;
  first: Boolean;
  N: Integer;
  S: string;
begin
  first := (writer.BaseStream.Size = 0);
  cnt := CountData(DataType);
  N := 0;
  DoProgress(0);
  while N < cnt do begin
    S := FetchData(DataType, cMaxCount, N + 1);
    if first and (N = 0) then begin
      S := ExtractFirstRows(S);
    end
    else begin
      S := ExtractRowData(S);
    end;
    N := N + cMaxCount;
    Writer.WriteLine(S);
    DoProgress(Round(100*N/cnt));
  end;
  DoProgress(100);
end;

procedure TQCScraper.Execute;
begin
  LoadAllData;
end;

end.
